// Author(s): Jan Friso Groote
// Copyright: see the accompanying file COPYING or copy at
// https://github.com/mCRL2org/mCRL2/blob/master/COPYING
//
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)
//

#ifndef MCRL2_ATERMPP_DETAIL_ATERM_H
#define MCRL2_ATERMPP_DETAIL_ATERM_H

#include "mcrl2/atermpp/detail/aterm_configuration.h"
#include "mcrl2/atermpp/detail/atypes.h"
#include "mcrl2/atermpp/function_symbol.h"
#include "mcrl2/utilities/noncopyable.h"
#include "mcrl2/utilities/shared_reference.h"
#include "mcrl2/atermpp/type_traits.h"

#include <limits>

namespace atermpp
{

// Forward declaration.
class unprotected_aterm;

constexpr static std::size_t MarkedReferenceCount = std::numeric_limits<std::size_t>::max();

namespace detail
{

/// \brief Can be used to check whether all elements in the parameter pack are derived from aterms.
template<typename ...Terms>
using are_terms = and_<std::is_convertible<Terms, unprotected_aterm>...>;

/// \brief This is the class to which an aterm points. Each _aterm consists
///        of a function symbol and a reference count used for garbage
///        collection.
class _aterm : public mcrl2::utilities::shared_reference_counted<_aterm, GlobalThreadSafe>, mcrl2::utilities::noncopyable
{
public:
  using ref = mcrl2::utilities::shared_reference<_aterm>;

  /// \brief Create a term from a function symbol.
  _aterm(const function_symbol& symbol) :
    m_function_symbol(symbol)
  {}

  function_symbol& function() noexcept
  {
    return m_function_symbol;
  }

  const function_symbol& function() const noexcept
  {
    return m_function_symbol;
  }

  /// \brief Mark this term to be garbage collected.
  /// \details Changes the reference count, so only apply whenever !is_reachable().
  void mark()
  {
    assert(!is_reachable());
    m_reference_count = MarkedReferenceCount;
    increment_reference_count_changes();
  }

  /// \brief Remove the mark from a term.
  /// \details Changes the reference count, so only apply whenever it was marked.
  void reset()
  {
    assert(is_marked());
    m_reference_count = 0;
    increment_reference_count_changes();
  }

  /// \brief During garbage collection a term will be marked whenever it occurs as the argument
  ///        of a reachable term, but is not protected itself.
  /// \returns True whenever this term has been marked.
  bool is_marked() const noexcept
  {
    return m_reference_count == MarkedReferenceCount;
  }

  /// \brief A term is reachable in the garbage collection graph if it is protected or whenever it occurs
  ///        as an argument of a reachable term. The latter will be ensured in the marking phase of garbage collection.
  /// \returns True whenever the term is reachable ie either marked or protected.
  bool is_reachable() const noexcept
  {
    return m_reference_count > 0;
  }

private:
  function_symbol m_function_symbol;
};

inline _aterm* address(const unprotected_aterm& t);

} // namespace detail
} // namespace atermpp

#endif /* MCRL2_ATERMPP_DETAIL_ATERM_H */
